package chaincode

import (
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"encoding/json"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
)

var (
	GlobalLogger = logging.Get()
)

var CommonChaincodeActionVar = map[string]CommonChaincodeAction{
	invoke.Ping: Ping, // Проверка установки чейнкода и возможности записи в леджер
	invoke.ClearLedger: ClearLedger, // Метод очищает все данные в ledger
}

var ChaincodeParameters ChaincodeParameter

var ChaincodeVar = CommonChaincode{
	Actions: CommonChaincodeActionVar,
}

func timeFuncFactory(adapter shim.ChaincodeStubInterface) func() (int64, error) {
	return func() (int64, error) {
		tt, err := adapter.GetTxTimestamp()
		if err != nil {
			return 0, err
		}
		return tt.Seconds, nil
	}
}

// Init initialize chaincode with proper parameters
func (t CommonChaincode) Init(adapter shim.ChaincodeStubInterface) (response pb.Response) {
	return shim.Success([]byte("Chaincode initialized successfully"))
}

func (t CommonChaincode) InitChaincode(initArgs []string) error {
	lenArguments := len(initArgs)
	if lenArguments == 1 {
		jsonChaincodeParameters := initArgs[0]

		GlobalLogger.Info("initArgs")
		GlobalLogger.Info(jsonChaincodeParameters)

		if err := json.Unmarshal([]byte(jsonChaincodeParameters), &ChaincodeParameters); err != nil {
			return GlobalLogger.Errorf(err.Error())
		}

		GlobalLogger.Info("ChaincodeParameters")
		GlobalLogger.Info(ChaincodeParameters.ToString())
	} else {
		return GlobalLogger.Errorf("Incorrect number of arguments. Expected 1 argument - ChaincodeParameter. But got %d", lenArguments)
	}

	return nil
}

// Invoke dispatch to the specific chaincode function based on the input parameters
func (t CommonChaincode) Invoke(adapter shim.ChaincodeStubInterface) (response pb.Response) {
	defer func() {
		GlobalLogger.Info("end init")
		if r := recover(); r != nil {
			errMsg := fmt.Sprintf("FATAL_ERROR: %f", r)
			response = shim.Error(errMsg)
			GlobalLogger.Error(errMsg)
		}
	}()
	functionName, args := adapter.GetFunctionAndParameters()
	GlobalLogger.Infof("ChannelName: %s", ChaincodeParameters.ChannelName)
	GlobalLogger.Infof("ChaincodeName: %s", ChaincodeParameters.ChaincodeName)
	GlobalLogger.Infof("functionName: %s", functionName)

	if functionName == "init" {
		defer func() {
			GlobalLogger.Info("end init")
			if r := recover(); r != nil {
				response = shim.Error(fmt.Sprintf("FATAL_ERROR: %f", r))
			}
		}()
		GlobalLogger.Info("Start initializing ..")

		if err := t.InitChaincode(args); err != nil {
			GlobalLogger.Error(err.Error())
			return shim.Error(err.Error())
		}

		return shim.Success([]byte("Chaincode initialized successfully"))
	}

	timeGetter := timeFuncFactory(adapter)
	_, err := timeGetter()
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(fmt.Sprintf("Failed to get timestamp, failed due to %s", err))
	}

	GlobalLogger.Info("ChaincodeParameters")
	GlobalLogger.Info(ChaincodeParameters.ToString())

	result, err := t.InvokeChaincode(functionName, args, adapter, timeGetter)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(err.Error())
	}

	bytes, err := json.Marshal(result)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return shim.Error(err.Error())
	}

	return shim.Success(bytes)
}

// Invoke dispatch to the specific chaincode function based on the input parameters
func (t CommonChaincode) InvokeChaincode(functionName string, args []string, adapter shim.ChaincodeStubInterface, timeGetter TimeGetter) (interface{}, error) {
	iManager := GetManager(adapter, timeGetter)
	userParameter, systemParameter, err := CheckAccessParameterByArgsAndFunctionName(ChaincodeParameters.ChannelName, ChaincodeParameters.ChaincodeName, args, functionName, iManager)
	if err != nil {
		GlobalLogger.Error(err.Error())
		return nil, err
	}

	if userParameter != nil {
		GlobalLogger.Infof("userParameter.UserID: %s", userParameter.UserID)
		GlobalLogger.Infof("userParameter.Param: %s", userParameter.Param)
	}

	if systemParameter != nil {
		GlobalLogger.Infof("systemParameter.UserID: %s", systemParameter.UserID)
		GlobalLogger.Infof("systemParameter.Param: %s", systemParameter.Param)
	}

	// taking care to dispatch to the correct function based on the input parameter
	if action, ok := t.Actions[functionName]; ok {
		return action(userParameter, systemParameter, iManager)
	}

	// No action define with provided function name, reporting error
	err = GlobalLogger.Errorf("Invalid invoke function name, function name provided %s ", functionName)
	return nil, err
}
