package userCC

import (
	"gitlab.smartblocklab.com/smartcontracts/common/security/actions"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_organization"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"github.com/pkg/errors"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"encoding/json"
	"reflect"
)

/**
	user - отдельный канал для работы с внешними серверами
		- user_id
		- public_key

	link_user - канал связей между пользователями внутри системы криптоентер
		кто кому роль

	link_user_status - канал статусов для связей между пользователями [link_user]
		user_id_user_id_status_time

	user_status - статус текущего пользователя по отношению к другому пользователю
		user_id user_id status статусы по отношению к кому то
 */
// User struct which specifies user record
type User struct {
	UserID       string                   `json:"user_id,omitempty"`
	PartnerCode  string                   `json:"partner_code,omitempty"`     // TODO уйдет в link_user
	CustomerType CustomerType             `json:"customer_type,omitempty"`    // TODO уйдет в link_user кто кому роль - статус связи выносится в другой канал link_user_status - user_id_user_id_status_time
	UserStatus   UserStatus               `json:"user_status,omitempty"`      // TODO вынесем в отдельный сервис - user_status - user_id user_id status статусы по отношению к кому то
	PublicKeys   map[string]UserPublicKey `json:"user_public_keys,omitempty"` // public_key Alias - unique | UserPublicKey
	Roles        []actions.Roles          `json:"roles"`                      // TODO уйдет в link_user
}

func (this User) CheckBeforeAdd(manager chaincode.IManager) error {
	err := VerifyPublicKey(manager, this)
	if err != nil {
		return err
	}

	if this.CustomerType == Root {
		return nil
		// PartnerCode need to be null or empty for RootUserID
	} else {
		organizationRemoteManager := remote_organization.GetOrganizationRemoteManager(manager, invoke.OrganizationChannel, invoke.OrganizationCC)
		organization, err := organizationRemoteManager.GetOrganization(&requests.Id{Id: this.PartnerCode})
		if err != nil {
			return errors.Errorf("Can't load organization. Error to execute method organizationCC.getOrganization with argument (partnerCode: %s) from organizationchannel",
				this.PartnerCode)
		}

		if organization == nil {
			return errors.Errorf("Organization with partner code %s not found", this.PartnerCode)
		}
	}
	return nil
}

func (this User) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	if this.UserID == system_user.ROOT_USER_ID {
		return errors.Errorf("You cannot update new root user")
	}

	for alias, publicKey := range this.PublicKeys {
		if publicKey.Alias != alias {
			return errors.Errorf("publicKey.Alias %s != alias %s", publicKey.Alias, alias)
		}
	}

	updatedUser := updatedObject.(User)

	// проверка соответствия новых ключей с существующими в блокчейне
	if len(updatedUser.PublicKeys) > 0 {
		found := false
		for updatedAlias, updatedPublicKey := range updatedUser.PublicKeys {
			found = false
			for searchAlias := range this.PublicKeys {
				if updatedAlias == searchAlias {
					found = true
					break
				}
			}

			if !found {
				return errors.Errorf("In updated public key array, public key not found by alias: %s.", updatedPublicKey.Alias)
			}
		}
	}

	return nil;
}

func (this User) New() chaincode.ValidObject {
	return &User{}
}

func (this User) Validate() error {
	errorMsg := ""

	if len(this.UserID) == 0 {
		errorMsg += "UserID is empty; "
	}
	if len(this.CustomerType) == 0 {
		errorMsg += "CustomerType is empty; "
	}
	if len(this.UserStatus) == 0 {
		errorMsg += "UserStatus is empty; "
	}
	if len(this.PublicKeys) == 0 {
		errorMsg += "PublicKeys is empty; "
	}

	if len(this.PartnerCode) != 0 && this.CustomerType == Root {
		// PartnerCode need to be null or empty for RootUserID
		errorMsg += fmt.Sprintf("Root user can't have PartnerCode '%s'; ", this.PartnerCode)
	} else if len(this.PartnerCode) == 0 && this.CustomerType != Root && this.CustomerType != System {
		errorMsg += "PartnerCode is empty; "
	}

	if !isExistsCustomerType(this.CustomerType) {
		errorMsg += fmt.Sprintf("CustomerType '%s' dose not exists; ", this.CustomerType)
	}

	if len(errorMsg) != 0 {
		return errors.New(errorMsg)
	}

	return nil
}

func isExistsCustomerType(CustomerType CustomerType) bool {
	if CustomerType == Person ||
		CustomerType == Company ||
		CustomerType == Root ||
		CustomerType == System ||
		CustomerType == Service {
		return true
	}
	return false
}

func (this User) GetKey(manager chaincode.IManager) []string {
	//TODO think abouth other key this.PartnerCode, this.UserID
	return []string{this.UserID}
}

func Unmarshal(data []byte) (*User, error) {
	var obj User
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *User) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*User, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check validObjectc is not reference")
	}

	casted, ok := validObject.(*User)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}

